#include "uart.h"
#include "calibration.h"
#include "RC5.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <stdint.h>
#include <stdlib.h>
  
//#define BAUD 2400
//#define BAUD 38400
#define BAUD 57600
//#define BAUD 115200



#define LED_RED PB0
#define LED_GREEN PB1
#define LED_BLUE PB2

// Watchdog reset handling

void early_wdt_reset(void) __attribute__((naked))  __attribute__((section(".init3")));
void early_wdt_reset(void) {
 // mcusr_mirror = MCUSR;
 MCUSR = 0;
  wdt_enable(WDTO_250MS);
// wdt_disable();
}
                                              


////////////////////////////////////////////////// BEGIN IRQ

uint8_t red_pwm,green_pwm,blue_pwm;

ISR(TIMER0_OVF_vect) {
  static uint8_t tick;

  TCNT0=0x7F - (tick>>1);

  if (tick < red_pwm)
    PORTB |= _BV(LED_RED);
  else
    PORTB &=~_BV(LED_RED);

  if (tick < green_pwm)
    PORTB |= _BV(LED_GREEN);
  else
    PORTB &=~_BV(LED_GREEN);

  if (tick < blue_pwm)
    PORTB |= _BV(LED_BLUE);
  else
    PORTB &=~_BV(LED_BLUE);


  tick++;
}

volatile uint8_t do_poll;
volatile uint8_t color_update;
ISR(TIMER1_OVF_vect) {
  static uint8_t color_divide;
  TCNT1 = -128;
  do_poll = 1;
  color_divide++;
  if (color_divide==0) {
    color_divide=0x7F;
    color_update++;
  }
}


////////////////////////////////////////////////// END IRQ

uint8_t H,S,V,Speed;

void hsv_to_rgb(void) {
	uint8_t r,g,b, i, f;
	uint8_t p, q, t;

	if( S == 0 ) 
 	{	r = g = b = V;
	}
	else
	{	i=H/43;
		f=H%43;
		p = (V * (255 - S))/256;
		q = (V * ((10710 - (S * f))/42))/256;
		t = (V * ((10710 - (S * (42 - f)))/42))/256;

		switch( i )
		{	case 0:
				r = V; g = t; b = p; break;
			case 1:
				r = q; g = V; b = p; break;
			case 2:
				r = p; g = V; b = t; break;
			case 3:
				r = p; g = q; b = V; break;			
			case 4:
				r = t; g = p; b = V; break;				
			case 5:
			default:
	 			r = V; g = p; b = q; break;
		}
	}

	r=get_calibration(red_calibration,r);
	g=get_calibration(green_calibration,g);
	b=get_calibration(blue_calibration,b);
	cli();                                              // Kein Interrupt beim Übernehmen der neuen Werte!
	red_pwm=r, green_pwm=g, blue_pwm=b;      
	sei();
}

/////// Predefined Colors

EEMEM uint8_t colors[10*4]={
// H, S, V, Speed
0x00,0x00,0xFF,0, // White
0x00,0xFF,0xFF,200, // Bright red
0x2A,0xFF,0xFF,200, // Bright Yellow
0x55,0xFF,0xFF,200, // Bright Green
0x80,0xFF,0xFF,200, // Bright Cyan
0xAA,0xFF,0xFF,200, // Bright Blue
0xD5,0xFF,0xFF,200, // Bright Viollet
0x00,0xFF,0xFF,10, // Fast Cycle
0x00,0x7F,0xFF,50, // Fast Cycle low Sat
0x00,0xFF,0x7F,50 // Fast Cycle low Value
};

////////////////////////

void rc5_handler(void) {
  uint8_t keycode;
  static uint8_t speed;
  static uint16_t last_data;
  uint8_t * eepos;
  
  if (rc5_data == last_data) {
    if (speed < 20)
      speed++;
  } else {
    speed=1;
    last_data = rc5_data;
  }
  
  keycode=rc5_data & 0xFF;

  if (keycode==0x20) {
    if (V <= 0xFF - speed) {
      V+=speed;
    } else {
      V=0xFF;
    }
  } else if (keycode==0x21) {
    if (V >= speed) {
      V-=speed;
    } else {
      V=0;
    }
  } else if (keycode==0x10) {
    if (S <= 0xFF - speed) {
      S+=speed;
    } else {
      S=0xFF;
    }
  } else if (keycode==0x11) {
    if (S >= speed) {
      S-=speed;
    } else {
      S=0;
    }
  } else if (keycode==0x0C) {
    if (Speed <= 0xFF - speed) {
      Speed+=speed;
    } else {
      Speed=0xFF;
    }
  } else if (keycode==0x0F) {
    if (Speed >= speed) {
      Speed-=speed;
    } else {
      Speed=0;
    }
  } else if (keycode==0x2E) {
    Speed=0xFF;
  } else if (keycode==0x0D) {
    V=0;
  } else if (keycode==0x22) {
    V=0xFF;
  } else if (keycode < 10) {
    eepos = &colors[keycode<<2];    
    H=eeprom_read_byte(eepos++);
    S=eeprom_read_byte(eepos++);
    V=eeprom_read_byte(eepos++);
    Speed=eeprom_read_byte(eepos++);
  }

  uart_write_P(PSTR("RC5: "));
  uart_hex(rc5_data>>8);
  uart_hex(rc5_data);
  uart_write_P(PSTR("\r\n"));
  rc5_data=0;
}

int main(void) {
  char c;

  DDRD = (1<<5);
  
  DDRB = _BV(LED_RED) | _BV(LED_GREEN) | _BV(LED_BLUE);
  uart_init(UART_BAUD_CALC(BAUD));


  TCCR0A = 0;
  TCCR0B = _BV(CS00); // F_OSC Timer Freq


  TCCR1A = 0;
  TCCR1B = _BV(CS11); // F_OSC / 8


  TIMSK = _BV(TOIE0) | _BV(TOIE1);
  

  H=0;
  S=255;
  V=255;
  Speed=175;
 
  hsv_to_rgb();  

  sei();

  uart_write_P(PSTR("OK\r\n")); 
  while (1) {
    wdt_reset();
    if (uart_recv_buf_size()) {
      c=uart_getc();
      uart_write_P(PSTR("H:"));
      uart_hex(H);
      uart_write_P(PSTR("\tS:"));
      uart_hex(S);
      uart_write_P(PSTR("\tV:"));
      uart_hex(V);
      uart_write_P(PSTR("\tR:"));
      uart_hex(red_pwm);
      uart_write_P(PSTR("\tG:"));
      uart_hex(green_pwm);
      uart_write_P(PSTR("\tB:"));
      uart_hex(blue_pwm);
      uart_write_P(PSTR("\r\n"));
    }
    if (do_poll) {
      do_poll=0;
      rc5_poll();
      if (rc5_data) {
        rc5_handler();
        hsv_to_rgb();  
      }
    } else if (Speed!=0xFF && (color_update > Speed)) { // "else": Prevent high update rate from messing with RC5...
      color_update=0;
      H++;
      hsv_to_rgb();  
    }

  } 
  return 0;
}

